#include "SteamTracking.h"



std::string SteamTracking::getDeviceTypeAsString(vr::ETrackedDeviceClass device_class)
{
	std::string type;
	switch (device_class)
	{
	case vr::TrackedDeviceClass_Controller:        type = "Controller"; break;
	case vr::TrackedDeviceClass_HMD:               type = "HMD"; break;
	case vr::TrackedDeviceClass_Invalid:           type = "Invalid"; break;
	case vr::TrackedDeviceClass_GenericTracker:    type = "GenericTracker"; break;
	case vr::TrackedDeviceClass_TrackingReference: type = "TrackingReference"; break;
	default:                                       type = "Unknown"; break;
	}
	return type;
}

glm::mat4x4* SteamTracking::convertMatrix(vr::HmdMatrix34_t in)
{
	/*
	return new glm::mat4x4(	in.m[0][0], in.m[0][1], in.m[0][2], in.m[0][3],
							in.m[1][0], in.m[1][1], in.m[1][2], in.m[1][3],
							in.m[2][0], in.m[2][1], in.m[2][2], in.m[2][3],
							0.0F,		0.0F,		 0.0F,		 1.0F);
	*/	
	return new glm::mat4x4(	in.m[0][0], in.m[1][0], in.m[2][0], 0.0F,
							in.m[0][1], in.m[1][1], in.m[2][1], 0.0F,
							in.m[0][2], in.m[1][2], in.m[2][2], 0.0F,
							in.m[0][3], in.m[1][3], in.m[2][3], 1.0F );
}



SteamTracking::SteamTracking()
{
	// Loading the SteamVR Runtime
	vr::EVRInitError eError = vr::VRInitError_None;
	vr_system_ = vr::VR_Init(&eError, vr::VRApplication_Overlay);

	if (eError != vr::VRInitError_None)
	{
		vr_system_ = nullptr;
		std::cerr <<"Unable to init VR runtime: "<< vr::VR_GetVRInitErrorAsEnglishDescription(eError) << std::endl;
	}

}


SteamTracking::~SteamTracking()
{
	if (vr_system_)
	{
		vr::VR_Shutdown();
		vr_system_ = nullptr;
	}
}




void SteamTracking::printVRDevices() {
	std::cout << "List of available VR devices:" << std::endl;
	if (!vr_system_) {
		std::cout << "\tNot connected to VR System, no devices"<< std::endl;
		return;
	}
	for (int nDevice = 0; nDevice < vr::k_unMaxTrackedDeviceCount; ++nDevice)
	{
		vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);
		if (device_class == vr::TrackedDeviceClass_Invalid)
		{
			continue;
		}

		const uint32_t size = 1024;
		char serial[size];
		vr_system_->GetStringTrackedDeviceProperty(nDevice, vr::ETrackedDeviceProperty::Prop_SerialNumber_String, serial, size);

		std::cout << "\tDevice " << nDevice << ":" << std::endl;
		std::cout << "\t\tType: " << getDeviceTypeAsString(device_class) << std::endl;
		std::cout << "\t\tSerial: " << serial << std::endl;
	}

	vr::HmdMatrix34_t raw_to_standing_htm = vr_system_->GetRawZeroPoseToStandingAbsoluteTrackingPose();
	std::cout << "RawZeroPoseToStandingAbsoluteTrackingPose:" << std::endl;
	printMatrix(raw_to_standing_htm);
}

void SteamTracking::processEvents()
{
	if (!vr_system_) return;

		vr::VREvent_t event;
		while (vr_system_->PollNextEvent(&event, sizeof(event)))
		{
			processEvent(event);
		}
}

int SteamTracking::getDeviceIDForSerial(char* Serial)
{
	if (!vr_system_) {
		return 1;
	}
	for (int nDevice = 0; nDevice < vr::k_unMaxTrackedDeviceCount; ++nDevice)
	{
		vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);
		if (device_class == vr::TrackedDeviceClass_Invalid)
		{
			continue;
		}

		const uint32_t size = 1024;
		char serial[size];
		vr_system_->GetStringTrackedDeviceProperty(nDevice, vr::ETrackedDeviceProperty::Prop_SerialNumber_String, serial, size);

		if (strcmp(Serial, serial) == 0) {
			std::cout << "Device " << Serial <<" has ID " << nDevice << std::endl;
			return nDevice;
		}
	}
	return 1;
}


void SteamTracking::printMatrix(const vr::HmdMatrix34_t &htm)
{
	for (int rows = 0; rows < 3; ++rows)
	{
		std::cout << (rows == 0 ? "[" : "") << "\t";

		for (int cols = 0; cols < 4; ++cols)
		{
			printf("%.3f\t", htm.m[rows][cols]);
			//std::cout << htm.m[rows][cols] << "\t";
		}
		std::cout << (rows == 2 ? "]" : "") << std::endl;
	}
}

void SteamTracking::processEvent(vr::VREvent_t event)
{
	switch (event.eventType)
	{
	case vr::VREvent_TrackedDeviceActivated:
		if (debugEvents) printf("Device %u attached.\n", event.trackedDeviceIndex);
		break;
	case vr::VREvent_TrackedDeviceDeactivated:
		if (debugEvents) printf("Device %u detached.\n", event.trackedDeviceIndex);
		break;
	case vr::VREvent_TrackedDeviceUpdated:
		if (debugEvents) printf("Device %u updated.\n", event.trackedDeviceIndex);
		break;
	case vr::VREvent_ButtonPress:
		if (onButtonPress) onButtonPress(event);
		if (debugEvents) printf("Device %u Button pressed %u.\n", event.trackedDeviceIndex, event.data.controller.button);
		break;
	case vr::VREvent_ButtonUnpress:
		if (debugEvents) printf("Device %u Button unpressed %u.\n", event.trackedDeviceIndex, event.data.controller.button);
		break;
	case vr::VREvent_ButtonTouch:
		if (debugEvents) printf("Device %u Button touched %u.\n", event.trackedDeviceIndex, event.data.controller.button);
		break;
	case vr::VREvent_ButtonUntouch:
		if (debugEvents) printf("Device %u Button untouched %u.\n", event.trackedDeviceIndex, event.data.controller.button);
		break;
	default:
		if (debugEvents) printf("Event: %u\n", event.eventType);
		break;
	}

}

glm::mat4x4* SteamTracking::getTransformationForDevice(int nDevice, float timeoffset)
{
	if (!vr_system_) {
		if (debug) std::cout << " NO VR System ";
		return new glm::mat4x4();
	}
	//TrackingUniverseStanding TrackingUniverseRawAndUncalibrated
	vr_system_->GetDeviceToAbsoluteTrackingPose(vr::TrackingUniverseStanding , timeoffset, tracked_device_poses_, vr::k_unMaxTrackedDeviceCount);

	if (!tracked_device_poses_[nDevice].bPoseIsValid)
	{
		if (debug) std::cout << " Device " << nDevice << "Pos is Invalid";
		return new glm::mat4x4();
	}

	vr::ETrackedDeviceClass device_class = vr_system_->GetTrackedDeviceClass(nDevice);

	if (device_class == vr::TrackedDeviceClass_Invalid)
	{
		if (debug) std::cout << " Device " << nDevice << "ist Invalid";
		return new glm::mat4x4();
	}

	vr::HmdMatrix34_t htm = tracked_device_poses_[nDevice].mDeviceToAbsoluteTracking;

	if (debug)
	{
		std::cout << "Pose of device " << nDevice << " (" << getDeviceTypeAsString(device_class)  << ")" << std::endl;
		printMatrix(htm);
	}

	return convertMatrix(htm);
}
